/*
 * We try to find an optimal triangle grid
 */

#include "common.h"
#include "minimise.h"
#include "mgraph.h"

static void compute_vertex_areas(const Vertices vertices, double areas[N]);
static double best_energy= DBL_MAX;

static void addcost(double *energy, double tweight, double tcost, int pr);
#define COST(weight, compute) addcost(&energy, (weight), (compute), printing)

/*---------- main energy computation and subroutines ----------*/

double compute_energy(const struct Vertices *vs) {
  double vertex_areas[N], energy;
  int printing;

  compute_vertex_areas(vs->a, vertex_areas);
  energy= 0;

  printing= printing_check(pr_cost);

  if (printing) printf("cost > energy |");

  COST(1e2, edgewise_vertex_displacement_cost(vs->a));
  COST(1e2, graph_layout_cost(vs->a,vertex_areas));
  COST(1e6, noncircular_rim_cost(vs->a));

  if (printing) printf("| total %# e |", energy);

  if (energy < best_energy) {
    FILE *best_f;
    int r;

    if (printing) printf(" BEST");

    best_f= fopen(output_file_tmp,"wb");  if (!best_f) diee("fopen new out");
    r= fwrite(vs->a,sizeof(vs->a),1,best_f); if (r!=1) diee("fwrite");
    if (fclose(best_f)) diee("fclose new best");
    if (rename(output_file_tmp,output_file)) diee("rename install new best");

    best_energy= energy;
  }
  if (printing) {
    putchar('\n');
    flushoutput();
  }

  return energy;
}

static void addcost(double *energy, double tweight, double tcost, int pr) {
  double tenergy= tweight * tcost;
  if (pr) printf(" %# e > %# e |", tcost, tenergy);
  *energy += tenergy;
}

static void compute_vertex_areas(const Vertices vertices, double areas[N]) {
  int v0,v1,v2, e1,e2, k;

  FOR_VERTEX(v0) {
    double total= 0.0;
    int count= 0;

    FOR_VEDGE(v0,e1,v1) {
      e2= (e1+1) % V6;
      v2= EDGE_END2(v0,e2);
      if (v2<0) continue;

      double e1v[D3], e2v[D3], av[D3];
      K {
	e1v[k]= vertices[v1][k] - vertices[v0][k];
	e2v[k]= vertices[v2][k] - vertices[v0][k];
      }
      xprod(av, e1v, e2v);
      total += magnD(av);
      count++;
    }
    areas[v0]= total / count;
  }
}

/*---------- Edgewise vertex displacement ----------*/

  /*
   *
   *
   *
   *   	            Q `-_
   *              / |    `-_
   *             /  |       `-.
   *            /   M - - - - - S
   *           /  ' |      _,-'
   *          /  '  |  _,-'
   *         / '  , P '
   *        / ',-'
   *       /,-'
   *      /'
   *   	 R
   *
   *  Let delta =  180deg - angle RMS
   *
   *  Let  l = |PQ|
   *       d = |RS|
   *
   *  Giving energy contribution:
   *
   *                                   3
   *                            l delta
   *    E             =  F   .  --------
   *     vd, edge PQ      vd       d
   *
   *
   *  (The dimensions of this are those of F_vd.)
   *
   *  We calculate delta as  atan2(|AxB|, A.B)
   *  where A = PQ, B = QR
   *
   *  In practice to avoid division by zero we'll add epsilon to d and
   *  |AxB| and the huge energy ought then to be sufficient for the
   *  model to avoid being close to R=S.
   */

double edgewise_vertex_displacement_cost(const Vertices vertices) {
  static const double axb_epsilon= 1e-6;

  int pi,e,qi,ri, k; //,si
  double  a[D3], b[D3], axb[D3]; //m[D3],
  double total_cost= 0;

  FOR_EDGE(qi,e,ri) {
    pi= EDGE_END2(qi,(e+3)%V6); if (pi<0) continue;

//    K m[k]= (vertices[pi][k] + vertices[qi][k]) * 0.5;
    K a[k]= -vertices[pi][k] + vertices[qi][k];
    K b[k]= -vertices[qi][k] + vertices[ri][k];

    xprod(axb,a,b);
    
    double delta= atan2(magnD(axb) + axb_epsilon, dotprod(a,b));
    double cost= pow(delta,3);

    if (!e && !(qi & YMASK))
      cost *= 10;

    total_cost += cost;
  }
  return total_cost;
}

/*---------- noncircular rim cost ----------*/

double noncircular_rim_cost(const Vertices vertices) {
  int vy,vx,v;
  double cost= 0.0;

  FOR_RIM_VERTEX(vy,vx,v) {
    double oncircle[3];
    /* By symmetry, nearest point on circle is the one with
     * the same angle subtended at the z axis. */
    oncircle[0]= vertices[v][0];
    oncircle[1]= vertices[v][1];
    oncircle[2]= 0;
    double mult= 1.0/ magnD(oncircle);
    oncircle[0] *= mult;
    oncircle[1] *= mult;
    double d2= hypotD2(vertices[v], oncircle);
    cost += d2*d2;
  }
  return cost;
}
